"""psy-strat: A psyplot plugin for stratigraphic plots

Describe your plugin here and do whatever you want. It should at least have a
__version__ attribute to specify the version of the package
"""

__version__ = "0.1.0"
